/*
http://www.philosophicalgeek.com/software/windows-media-top-10-plug-in/
Copyright (c) 2007 Ben Watson (http://www.philosophicalgeek.com)

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this software.

Permission is granted to anyone to use this software for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
not claim that you wrote the original software. If you use this
software in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
distribution.
*/
#include "StdAfx.h"
#include "Top10Database.h"
#include "shlobj.h"

const BYTE CTop10Database::m_nDBVersion = 1;
const int CTop10Database::m_nValueLengthLimit = 256;//longest value that can be written to database

CTop10Database::CTop10Database(void)
{
}

CTop10Database::~CTop10Database(void)
{
}

/*
AddSong takes an array of CString vectors. Each array represents an attribute,
with the vector containing 0 or more values belonging to that attribute
*/
void CTop10Database::AddSong(const vector<CString> strValues[saNumSongAttributes])
{
	for (int i=0;i<saNumSongAttributes;i++)
	{
		const vector<CString>& rValues = strValues[i];
		size_t numValues = rValues.size();
		for (unsigned int v=0;v<numValues;v++)
		{
			const CString& rValue = rValues.at(v);
			if (rValue.GetLength() > 0)
			{
				const CValueCountPair& pair = m_masterList[i].AddValue(rValue);
				//add it to top 10 list
				m_topLists[i].Add(pair);
				SetDirty();
			}
		}

	}

}

//returns the value/count pair from the given category/index
CValueCountPair& CTop10Database::GetTopValue(SongAttribute nCategory, int nIndex)
{
	if (nCategory >=0 && nCategory < saNumSongAttributes)
	{
		if (nIndex>=0 && nIndex<GetNumValues(nCategory))
		{
			return m_topLists[nCategory].GetAt(nIndex);
		}
		
	}
	
	_ASSERT(0);
	
}

//wipes playback records clean
void CTop10Database::Clear()
{
	for (int i=0;i<saNumSongAttributes;i++)
	{
		m_topLists[i].Clear();
		m_masterList[i].Clear();
	}
	SetDirty();
}

//returns the number of values in a category
int CTop10Database::GetNumValues(SongAttribute nCategory) const 
{
	if (nCategory >=0 && nCategory < saNumSongAttributes)
	{
		return m_topLists[nCategory].NumItems();
	}
	_ASSERT(0);
	return 0;
}

bool CTop10Database::Load()
{
	CString strFilename = GetDBFilename();
	HANDLE hFile = CreateFile(strFilename, 
		GENERIC_READ,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_SEQUENTIAL_SCAN,
		NULL);

	if (hFile==INVALID_HANDLE_VALUE)
		return false;

	//read first byte, which is version number of database
	DWORD dwBytesRead = 0;

	BYTE nVer;
	BOOL bSuccess = ReadFile(hFile,&nVer,sizeof(BYTE),&dwBytesRead, NULL);
	if (!bSuccess || dwBytesRead!=sizeof(BYTE))
	{
		CloseHandle(hFile);
		return false;
	}

	bool bLoaded = false;

	
	switch (nVer)
	{
	case 1:
		bLoaded = LoadVersion1(hFile);

		break;
	default:
		_ASSERT(0);
		break;
	}


	CloseHandle(hFile);

	ClearDirty();

	
	return bLoaded;
}

bool CTop10Database::LoadVersion1(HANDLE hFile)
{
	//read in number of sections
	DWORD dwNumSections = 0;
	DWORD dwBytesRead = 0;

	BOOL bSuccess = ReadFile(hFile,&dwNumSections,sizeof(dwNumSections),&dwBytesRead, NULL);
	if (!bSuccess || dwBytesRead!=sizeof(dwNumSections))
	{
		_ASSERT(0);
		return false;
	}
	//sanity check
	if (dwNumSections<0 || dwNumSections>saNumSongAttributes)
	{
		_ASSERT(0);
		return false;
	}

	for (unsigned int iSection = 0;iSection < dwNumSections;iSection++)
	{
		//read section header
		DBSECTIONHEADERSTRUCT sectionHeader;
		bSuccess = ReadFile(hFile, &sectionHeader,sizeof(sectionHeader),&dwBytesRead,NULL);
		if (!bSuccess || dwBytesRead!=sizeof(sectionHeader))
		{
			_ASSERT(0);
			return false;
		}

		//sanity check
		if (sectionHeader.nSectionId<0 || sectionHeader.nSectionId >=saNumSongAttributes)
		{
			_ASSERT(0);
			return false;
		}

		if (sectionHeader.dwNumTop10Values < 0 || sectionHeader.dwNumValues < 0)
		{
			_ASSERT(0);
			return false;
		}

		//read in top-10 values
		for (unsigned int iSong = 0;iSong < sectionHeader.dwNumTop10Values; iSong++)
		{
			CValueCountPair pair;
			bSuccess = ReadValueCountPair(hFile, pair);
			if (!bSuccess)
			{
				_ASSERT(0);
				return false;
			}
			m_topLists[iSection].Add(pair);
		}

		//read in all values
		for (unsigned int iSong = 0;iSong < sectionHeader.dwNumValues;iSong++)
		{
			CValueCountPair pair;
			bSuccess = ReadValueCountPair(hFile, pair);
			if (!bSuccess)
			{
				_ASSERT(0);
				return false;
			}
			m_masterList[iSection].AddValue(pair);
		}
	}

	return true;
}

bool CTop10Database::Save() 
{
	if (!IsDirty())
		return true;

	CString strFilename = GetDBFilename();

	HANDLE hFile = CreateFile(strFilename, 
		GENERIC_WRITE,
		0,
		NULL,
		CREATE_ALWAYS,
		FILE_FLAG_SEQUENTIAL_SCAN,
		NULL);

	//always save as latest version
	if (hFile==INVALID_HANDLE_VALUE)
	{
#ifdef _DEBUG
		int nError = GetLastError();
		printf("Last Error: %d",nError);
#endif
		return false;
	}
	
	//database header
	DBHEADERSTRUCT header;
	header.nVersion = m_nDBVersion;
	header.dwNumSections = saNumSongAttributes;

	DWORD dwBytesWritten = 0;
	BOOL bSuccess = WriteFile(hFile, &header,sizeof(header),&dwBytesWritten, NULL);
	if (dwBytesWritten < sizeof(header) || !bSuccess) {
		CloseHandle(hFile);
		return false;
	}

	//write out top 10 and full lists for each section
	for (int i=0;i<saNumSongAttributes;i++)
	{
		DBSECTIONHEADERSTRUCT sectionHeader;
		sectionHeader.nSectionId = i;
		sectionHeader.dwNumTop10Values = (DWORD)m_topLists[i].NumItems();
		sectionHeader.dwNumValues = (DWORD)m_masterList[i].NumItems();		

		bSuccess = WriteFile(hFile, &sectionHeader, sizeof(sectionHeader), &dwBytesWritten, NULL);
		if (dwBytesWritten < sizeof(sectionHeader) || !bSuccess)
		{
			CloseHandle(hFile);
			return false;
		}

		//write out top-10 values for current section
		for (unsigned int v = 0;v < sectionHeader.dwNumTop10Values; v++)
		{
			const CValueCountPair& pair = m_topLists[i].GetAt(v);
			if (!WriteValueCountPair(hFile, pair))
			{
				CloseHandle(hFile);
				return false;
			}
		}

		//write out all values from master list for the current section
		
		DWORD dwNumWritten = 0;
		CValueCountsIterator iter = m_masterList[i].GetIterator();
		CValueCountPair pair;
		iter.First();
		while (iter.Next(pair))
		{
			if (!WriteValueCountPair(hFile, pair))
			{
				CloseHandle(hFile);
				return false;
			}
			dwNumWritten++;
		}
		_ASSERT(dwNumWritten == sectionHeader.dwNumValues);
		
	}

	

	CloseHandle(hFile);

	ClearDirty();

	return true;
}
////////////////////////////////////////////////////////////////////////////////////////
// Helper functions for reading/writing database values
//
////////////////////////////////////////////////////////////////////////////////////////


bool CTop10Database::ReadValueCountPair(HANDLE hFile, CValueCountPair& rPair) const
{
	_ASSERT(hFile);
	DBVALUESTRUCT valueStruct;
	DWORD dwBytesRead = 0;

	BOOL bSuccess = ReadFile(hFile,&valueStruct,sizeof(valueStruct),&dwBytesRead, NULL);
	if (!bSuccess || dwBytesRead!=sizeof(valueStruct))
	{
		_ASSERT(0);
		return false;
	}

	//sanity check
	if (valueStruct.cchValueLength < 0 || valueStruct.cchValueLength>m_nValueLengthLimit)
	{
		_ASSERT(0);
		return false;
	}

	if (valueStruct.dwCount < 0) 
	{
		_ASSERT(0);
		return false;
	}

	
#ifdef _DEBUG
	//make sure our string is of the wide-character variety
	const WCHAR* pChar = rPair.Value();
#endif
	
	bSuccess = ReadFile(hFile, rPair.Value().GetBufferSetLength(valueStruct.cchValueLength),valueStruct.cchValueLength * sizeof(WCHAR),&dwBytesRead,NULL);
	rPair.Value().ReleaseBuffer();

	if (!bSuccess || dwBytesRead != valueStruct.cchValueLength * sizeof(WCHAR))
	{
		_ASSERT(0);
		
		return false;
	}
	rPair.Count(valueStruct.dwCount);

	return true;
}

bool CTop10Database::WriteValueCountPair(HANDLE hFile, const CValueCountPair& rPair) const
{
	_ASSERT(hFile);
	//write value length, value, then count
	DBVALUESTRUCT valueStruct;
	valueStruct.cchValueLength = rPair.Value().GetLength();
	valueStruct.dwCount = rPair.Count();
	const WCHAR* pBuffer = rPair.Value();
	
	DWORD dwBytesWritten = 0;
	BOOL bSuccess = WriteFile(hFile, &valueStruct,sizeof(valueStruct),&dwBytesWritten, NULL);
	if (!bSuccess || dwBytesWritten < sizeof(valueStruct))
	{
		return false;
	}

	//limit value length for sanity (should always be equal)
	int nLength = min(valueStruct.cchValueLength, m_nValueLengthLimit);
	_ASSERT(nLength == valueStruct.cchValueLength);	

	bSuccess = WriteFile(hFile, pBuffer, 
		nLength * sizeof(pBuffer[0]), 
		&dwBytesWritten,
		NULL);

	if (!bSuccess || dwBytesWritten < sizeof(pBuffer[0]) * nLength)
	{
		return false;
	}
	
	return true;
}


CString CTop10Database::GetDBFilename(void) const
{
	//store my data in <User\AppData>\WindowsMediaTop10

	CString strDataPath = _T("\\WindowsMediaTop10");
	CString strFileName = _T("\\WMTop10DB.dat");
		
	CString strPath;
	HRESULT hResult = SHGetFolderPath(
		NULL,
		CSIDL_APPDATA,
		NULL,
		SHGFP_TYPE_CURRENT,
		strPath.GetBufferSetLength(MAX_PATH));

	strPath.ReleaseBuffer();

	if (SUCCEEDED(hResult))
	{
		strPath+=strDataPath;
		//create path if it doesn't exist
		::CreateDirectory(strPath, NULL);

		return strPath + strFileName;
	}

	_ASSERT(0);
	return CString();
}


